/*
 * Copyright 2013-2015 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.springframework.cloud.sleuth.instrument.web;

import java.util.Random;
import java.util.regex.Pattern;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.junit.Before;
import org.junit.Test;
import org.mockito.Mock;
import org.springframework.cloud.sleuth.DefaultSpanNamer;
import org.springframework.cloud.sleuth.Sampler;
import org.springframework.cloud.sleuth.Span;
import org.springframework.cloud.sleuth.SpanExtractor;
import org.springframework.cloud.sleuth.SpanInjector;
import org.springframework.cloud.sleuth.SpanReporter;
import org.springframework.cloud.sleuth.TraceKeys;
import org.springframework.cloud.sleuth.Tracer;
import org.springframework.cloud.sleuth.log.SpanLogger;
import org.springframework.cloud.sleuth.sampler.AlwaysSampler;
import org.springframework.cloud.sleuth.sampler.NeverSampler;
import org.springframework.cloud.sleuth.trace.DefaultTracer;
import org.springframework.cloud.sleuth.trace.TestSpanContextHolder;
import org.springframework.http.HttpStatus;
import org.springframework.http.MediaType;
import org.springframework.mock.web.MockFilterChain;
import org.springframework.mock.web.MockHttpServletRequest;
import org.springframework.mock.web.MockHttpServletResponse;
import org.springframework.mock.web.MockServletContext;
import org.springframework.test.web.servlet.request.MockHttpServletRequestBuilder;

import static org.assertj.core.api.Assertions.assertThat;
import static org.assertj.core.api.Assertions.entry;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNull;
import static org.mockito.MockitoAnnotations.initMocks;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.get;

/**
 * @author Spencer Gibb
 */
public class TraceFilterTests {

	@Mock SpanLogger spanLogger;
	@Mock SpanReporter spanReporter;
	SpanExtractor<HttpServletRequest> spanExtractor = new HttpServletRequestExtractor(new Random(), Pattern
			.compile(TraceFilter.DEFAULT_SKIP_PATTERN));
	SpanInjector<HttpServletResponse> spanInjector = new HttpServletResponseInjector();

	private Tracer tracer;
	private TraceKeys traceKeys = new TraceKeys();

	private Span span;

	private MockHttpServletRequest request;
	private MockHttpServletResponse response;
	private MockFilterChain filterChain;
	private Sampler sampler = new AlwaysSampler();

	@Before
	public void init() {
		initMocks(this);
		this.tracer = new DefaultTracer(new DelegateSampler(), new Random(),
				new DefaultSpanNamer(), this.spanLogger, this.spanReporter) {
			@Override
			public Span continueSpan(Span span) {
				TraceFilterTests.this.span = super.continueSpan(span);
				return TraceFilterTests.this.span;
			}
		};
		this.request = builder().buildRequest(new MockServletContext());
		this.response = new MockHttpServletResponse();
		this.response.setContentType(MediaType.APPLICATION_JSON_VALUE);
		this.filterChain = new MockFilterChain();
	}

	public MockHttpServletRequestBuilder builder() {
		return get("/?foo=bar").accept(MediaType.APPLICATION_JSON).header("User-Agent",
				"MockMvc");
	}

	@Test
	public void notTraced() throws Exception {
		this.sampler = NeverSampler.INSTANCE;
		TraceFilter filter = new TraceFilter(this.tracer, this.traceKeys, this.spanReporter,
				this.spanExtractor, this.spanInjector);

		this.request = get("/favicon.ico").accept(MediaType.ALL)
				.buildRequest(new MockServletContext());

		filter.doFilter(this.request, this.response, this.filterChain);

		assertFalse(this.span.isExportable());
		assertNull(TestSpanContextHolder.getCurrentSpan());
	}

	@Test
	public void startsNewTrace() throws Exception {
		TraceFilter filter = new TraceFilter(this.tracer, this.traceKeys, this.spanReporter,
				this.spanExtractor, this.spanInjector);
		filter.doFilter(this.request, this.response, this.filterChain);
		verifyHttpTags();
		assertNull(TestSpanContextHolder.getCurrentSpan());
	}

	@Test
	public void startsNewTraceWithParentIdInHeaders() throws Exception {
		this.request = builder()
				.header(Span.SPAN_ID_NAME, Span.idToHex(1L))
				.header(Span.TRACE_ID_NAME, Span.idToHex(2L))
				.header(Span.PARENT_ID_NAME, Span.idToHex(3L))
				.buildRequest(new MockServletContext());
		TraceFilter filter = new TraceFilter(this.tracer, this.traceKeys, this.spanReporter,
				this.spanExtractor, this.spanInjector);

		filter.doFilter(this.request, this.response, this.filterChain);

		// this creates a child span which is why we'd expect the parents to include 1L)
		assertThat(this.span.getParents()).containsOnly(1L);
		assertNull(TestSpanContextHolder.getCurrentSpan());
	}

	@Test
	public void continuesSpanInRequestAttr() throws Exception {
		Span span = this.tracer.createSpan("http:foo");
		this.request.setAttribute(TraceFilter.TRACE_REQUEST_ATTR, span);
		// It should have been removed from the thread local context so simulate that
		TestSpanContextHolder.removeCurrentSpan();

		TraceFilter filter = new TraceFilter(this.tracer, this.traceKeys, this.spanReporter,
				this.spanExtractor, this.spanInjector);
		filter.doFilter(this.request, this.response, this.filterChain);

		verifyHttpTags();

		assertNull(TestSpanContextHolder.getCurrentSpan());
	}

	@Test
	public void continuesSpanFromHeaders() throws Exception {
		this.request = builder().header(Span.SPAN_ID_NAME, 10L)
				.header(Span.TRACE_ID_NAME, 20L).buildRequest(new MockServletContext());

		TraceFilter filter = new TraceFilter(this.tracer, this.traceKeys, this.spanReporter,
				this.spanExtractor, this.spanInjector);
		filter.doFilter(this.request, this.response, this.filterChain);

		verifyHttpTags();

		assertNull(TestSpanContextHolder.getCurrentSpan());
	}

	@Test
	public void addsAdditionalHeaders() throws Exception {
		this.request = builder().header(Span.SPAN_ID_NAME, 10L)
				.header(Span.TRACE_ID_NAME, 20L).buildRequest(new MockServletContext());

		this.traceKeys.getHttp().getHeaders().add("x-foo");
		TraceFilter filter = new TraceFilter(this.tracer, this.traceKeys, this.spanReporter,
				this.spanExtractor, this.spanInjector);
		this.request.addHeader("X-Foo", "bar");
		filter.doFilter(this.request, this.response, this.filterChain);

		assertThat(this.span.tags()).contains(entry("http.x-foo", "bar"));

		assertNull(TestSpanContextHolder.getCurrentSpan());
	}

	@Test
	public void ensuresThatParentSpanIsStoppedWhenReported() throws Exception {
		this.request = builder().header(Span.SPAN_ID_NAME, 10L)
				.header(Span.TRACE_ID_NAME, 20L).buildRequest(new MockServletContext());
		TraceFilter filter = new TraceFilter(this.tracer, this.traceKeys, spanIsStoppedVeryfingReporter(),
				this.spanExtractor, this.spanInjector);

		filter.doFilter(this.request, this.response, this.filterChain);
	}

	SpanReporter spanIsStoppedVeryfingReporter() {
		return (span) -> assertThat(span.getEnd()).as("Span has to be stopped before reporting").isNotZero();
	}

	@Test
	public void additionalMultiValuedHeader() throws Exception {
		this.request = builder().header(Span.SPAN_ID_NAME, 10L)
				.header(Span.TRACE_ID_NAME, 20L).buildRequest(new MockServletContext());

		this.traceKeys.getHttp().getHeaders().add("x-foo");
		TraceFilter filter = new TraceFilter(this.tracer, this.traceKeys, this.spanReporter,
				this.spanExtractor, this.spanInjector);
		this.request.addHeader("X-Foo", "bar");
		this.request.addHeader("X-Foo", "spam");
		filter.doFilter(this.request, this.response, this.filterChain);

		assertThat(this.span.tags()).contains(entry("http.x-foo", "'bar','spam'"));

		assertNull(TestSpanContextHolder.getCurrentSpan());
	}

	@Test
	public void catchesException() throws Exception {
		TraceFilter filter = new TraceFilter(this.tracer, this.traceKeys, this.spanReporter,
				this.spanExtractor, this.spanInjector);
		this.filterChain = new MockFilterChain() {
			@Override
			public void doFilter(javax.servlet.ServletRequest request,
					javax.servlet.ServletResponse response)
							throws java.io.IOException, javax.servlet.ServletException {
				throw new RuntimeException("Planned");
			};
		};
		try {
			filter.doFilter(this.request, this.response, this.filterChain);
		}
		catch (RuntimeException e) {
			assertEquals("Planned", e.getMessage());
		}
		verifyHttpTags(HttpStatus.INTERNAL_SERVER_ERROR);

		assertNull(TestSpanContextHolder.getCurrentSpan());
	}

	public void verifyHttpTags() {
		verifyHttpTags(HttpStatus.OK);
	}

	/**
	 * Shows the expansion of {@link import
	 * org.springframework.cloud.sleuth.instrument.TraceKeys}.
	 */
	public void verifyHttpTags(HttpStatus status) {
		assertThat(this.span.tags()).contains(entry("http.host", "localhost"),
				entry("http.url", "http://localhost/?foo=bar"), entry("http.path", "/"),
				entry("http.method", "GET"));

		// Status is only interesting in non-success case. Omitting it saves at least
		// 20bytes per span.
		if (status.is2xxSuccessful()) {
			assertThat(this.span.tags()).doesNotContainKey("http.status_code");
		}
		else {
			assertThat(this.span.tags()).containsEntry("http.status_code",
					status.toString());
		}
	}

	private class DelegateSampler implements Sampler {
		@Override
		public boolean isSampled(Span span) {
			return TraceFilterTests.this.sampler.isSampled(span);
		}
	}
}
